#!/usr/bin/perl
###########################################################################
# lsb-functions-signature-checker v1.0, test suite for checking ABI compliance
# of two Linux distributions.
# Copyright (C) The Linux Foundation
# Copyright (C) Institute for System Programming, RAS
# Author: Andrey Ponomarenko
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
###########################################################################
use Getopt::Long;
use FindBin;
sub BEGIN {
 unshift @INC, $FindBin::Bin;
}
use SimpleJournal;
Getopt::Long::Configure ("posix_default", "no_ignore_case");

sub BEGIN {
 unshift @INC, $FindBin::Bin;
}

my ($Help, $Run_Tests, $Create_TestSuite, $Distribution_Name, $Source, $TargetLibrary);

GetOptions("help|h!" => \$Help,
 "distr=s" => \$Distribution_Name,
 "src=s" => \$Source,
 "create!" => \$Create_TestSuite,
 "run!" => \$Run_Tests,
 "l=s" => \$TargetLibrary);

sub HELP_MESSAGE()
{
    print STDERR <<"EOM"

Name:
  $0: test suite for checking ABI compliance of two Linux distributions.

Usage:
  $0 [options] 

Options:
  -help
      This help message.

  -distr
      The name of distribution.

  -src <path>
      Path to the test suite source.

  -create
      Create test suite (on first distribution).

  -run
      Run tests (on second distribution).

  -l <library>
      This option allow to recreate test case for the specified library.

Examples of Use:
  $0 -create -src moblin-functions-signature-checker-data -distr Moblin
  $0 -run -src moblin-functions-signature-checker-data -distr Ubuntu

EOM
      ;
      exit(0);
}

my %TestResult;
my %Descriptor;
my %DevelFiles;
my $TESTSUITE_DISTRIBUTION;
my $TOOL_DIR;

sub recreate_test_case($)
{
    my $Library = $_[0];
    my $D_Path = $Source."/descriptors/$Library\.desc";
    my $Int_List_Path = $Source."/int_list/$Library\.ints";
    print "dumping $Library ABI\n";
    print "running abi-compliance-checker.pl -l \'$Library\' -dump_info \'$D_Path\' -int_list \'$Int_List_Path\'\n";
    system("$FindBin::Bin/abi-compliance-checker.pl -l \'$Library\' -dump_info \'$D_Path\' -int_list \'$Int_List_Path\'");
    print "\n";
    system("mkdir", "-p", $Source."/test_data/".$Library);
    system("mv -f abi_dumps/$Library/* ".esc($Source)."/test_data/$Library/");
    `rm -fr abi_dumps/`;
}

sub esc($)
{
    my $Str = $_[0];
    $Str =~ s/([()\[\]{}$ &'"`;])/\\$1/g;
    return $Str;
}

sub cmd_find($$$)
{
    my ($Path, $Type, $Name) = @_;
    return () if(not $Path or not -e $Path);
    my $Cmd = "find ".esc($Path);
    if($Type)
    {
        $Cmd .= " -type $Type";
    }
    if($Name)
    {
        $Cmd .= " -name \"$Name\"";
    }
    return split("\n", `$Cmd`);
}

sub create_test_suite()
{#Dump ABI information from Moblin library descriptors
    if(-e "abi_dumps")
    {
        `rm -fr abi_dumps`;
    }
    foreach my $D_Path (sort (cmd_find($Source."/descriptors","f","")))
    {
        $D_Path =~ /\/([^\/]*)\.desc\Z/;
        my $Library = $1;
        my $Int_List_Path = $Source."/int_list/".$Library.".ints";
        print "dumping $Library ABI\n";
        print "running abi-compliance-checker.pl -l \'$Library\' -dump_info \'$D_Path\' -int_list \'$Int_List_Path\'\n";
        system("$FindBin::Bin/abi-compliance-checker.pl -l \'$Library\' -dump_info \'$D_Path\' -int_list \'$Int_List_Path\'");
        print "\n";
    }
    if(-e $Source."/test_data/")
    {
        system("rm", "-fr", $Source."/test_data/");
    }
    system("mv", "abi_dumps", $Source."/test_data");
    system("echo \"$Distribution_Name\" > ".esc($Source)."/distribution.name");
    print "test data for $Distribution_Name has been created in the directory '$Source/test_data/'\n";
}

sub run_tests()
{#Run abi-compliance-checker on two descriptors sets
    `rm -fr compat_reports target_descriptors results`;
    `mkdir -p results`;
    my @Descriptors = cmd_find($Source."/descriptors","f","");
    my @ABIs = cmd_find($Source."/test_data","f","");
    my $JOURNAL = SimpleJournal->new("results/journal");
    $JOURNAL->test_name("ABI compliance test for $Distribution_Name and $TESTSUITE_DISTRIBUTION");
    $JOURNAL->tests_expected($#ABIs+1);
    foreach my $D_Path (sort {lc($a) cmp lc($b)} @Descriptors)
    {
        $D_Path =~ /\/([^\/]*)\.desc\Z/;
        my $Library = $1;
        next if(not $Library or (defined $TargetLibrary and $Library ne $TargetLibrary));
        print "testing for $Library ABI compliance\n";
        $JOURNAL->test_case_start($Library);
        print "creating target descriptor for library $Library\n";
        my $ABI_Path = (cmd_find($Source."/test_data/$Library","f","$Library\_*abi\.tar\.gz"))[0];
        next if(not $ABI_Path);
        if(create_target_descriptor($Library, $D_Path))
        {
            my $Int_List_Path = $Source."/int_list/".$Library.".ints";
            print "running abi-compliance-checker.pl -l \'$Library\' -d1 \'target_descriptors/$Library\.desc\' -d2 \'$ABI_Path\' -int_list \'$Int_List_Path\'\n";
            system("$FindBin::Bin/abi-compliance-checker.pl -l \'$Library\' -d1 \'target_descriptors/$Library\.desc\' -d2 \'$ABI_Path\' -int_list \'$Int_List_Path\'");
            print "\n";
            check_result($Library);
            my $Report_Path = (cmd_find("compat_reports/$Library","f","abi_compat_report.html"))[0];
            $JOURNAL->test_description("{noreplace}Test for library <a href=\'$Report_Path\'>$Library ABI compliance</a>{/noreplace}");
            my $Result = $TestResult{$Library}{"Withdrawn"}+$TestResult{$Library}{"T_High"}+$TestResult{$Library}{"T_Medium"}+$TestResult{$Library}{"I_High"}+$TestResult{$Library}{"I_Medium"};
            if($Result==0)
            {
                $JOURNAL->test_case_result("passed");
            }
            else
            {
                $JOURNAL->test_comment("{noreplace}".create_summary($TestResult{$Library}{"Withdrawn"}, $TestResult{$Library}{"T_High"}, $TestResult{$Library}{"T_Medium"}, $TestResult{$Library}{"T_Low"}, $TestResult{$Library}{"I_High"}, $TestResult{$Library}{"I_Medium"}, $TestResult{$Library}{"I_Low"}, $TestResult{$Library}{"Constants"})."{/noreplace}");
                $JOURNAL->test_case_result("failed");
            }
            $TestResult{$Library}{"IsCorrect"} = 1;
        }
        else
        {
            $JOURNAL->test_description("Test for library $Library ABI compliance");
            print "can't create target descriptor for library $Library\n";
            $JOURNAL->test_comment("necessary header files and shared objects for $Library were not found");
            $JOURNAL->test_case_result("failed");
            $TestResult{$Library}{"IsCorrect"} = 0;
        }
        print "\n";
    }
    $JOURNAL->close();
}

sub word_tr($$)
{
    my ($word, $count) = @_;
    if($count == 1)
    {
        return $word;
    }
    else
    {
        return $word."s";
    }
}

sub create_index()
{
    my $CssStyles = "<style type=\"text/css\">
    hr{color:Black;background-color:Black;height:1px;border: 0;}
    td.table_header{background-color:#eeeeee;}
    td.summary_item{font-size:15px;font-family:Arial;}
    td.summary_item_value{padding-right:5px;width:50px;}
    a.report_link{color:Blue;font-size:12px;}
    td.warnings{background-color:#F4F4AF;}
    td.failures{background-color:#FFC3CE;}
    td.success{background-color:#CCFFCC;}</style>\n";
    my $LIBS_INFO = "<h2 style='margin-bottom:0px;padding-bottom:0px;'>Details for Libraries (".keys(%TestResult).")</h2><hr/>\n<table cellpadding='3' border='1' style='border-collapse:collapse;'>";
    $LIBS_INFO .= "<tr><td rowspan='2' class='table_header'><b>Library</b></td><td rowspan='2' class='table_header' align='center'><b>Absent Interfaces<br/>in $TESTSUITE_DISTRIBUTION</b></td><td colspan='3' class='table_header' align='center'><b>Problems in Data Types</b></td><td colspan='3' class='table_header' align='center'><b>Interface Problems</b></td><td rowspan='2' class='table_header' align='center'><b>Constants<br/>Problems</b></td></tr>\n<tr><td class='table_header' style='color:Red;' align='center'>High risk</td><td class='table_header' align='center'>Medium risk</td><td class='table_header' align='center'>Warnings</td><td class='table_header' style='color:Red;' align='center'>High risk</td><td class='table_header' align='center'>Medium risk</td><td class='table_header' align='center'>Warnings</td></tr>\n";
    my ($Total_Libs, $Total_Miss) = (0, 0);
    my ($Withdrawn_Total, $T_High_Total, $T_Medium_Total, $T_Low_Total, $I_High_Total, $I_Medium_Total, $I_Low_Total, $Constants_Total) = (0, 0, 0, 0, 0, 0, 0, 0);
    foreach my $Library (sort {lc($a) cmp lc($b)} keys(%TestResult))
    {
        if(not $TestResult{$Library}{"IsCorrect"})
        {
            $Total_Miss += 1;
            $LIBS_INFO .= "<tr><td><a>$Library</a></td><td colspan='8' style='color:Red;'>&nbsp;&nbsp;&nbsp;necessary header files and shared objects were not found</td></tr>\n";
            next;
        }
        next if(not -e "compat_reports/$Library");
        $Total_Libs += 1;
        $Withdrawn_Total += $TestResult{$Library}{'Withdrawn'};
        $T_High_Total += $TestResult{$Library}{'T_High'};
        $T_Medium_Total += $TestResult{$Library}{'T_Medium'};
        $T_Low_Total += $TestResult{$Library}{'T_Low'};
        $I_High_Total += $TestResult{$Library}{'I_High'};
        $I_Medium_Total += $TestResult{$Library}{'I_Medium'};
        $I_Low_Total += $TestResult{$Library}{'I_Low'};
        $Constants_Total += $TestResult{$Library}{'Constants'};
        
        my $Report_Path = (cmd_find("compat_reports/$Library","f","abi_compat_report.html"))[0];
        
        my $Withdrawn_Link = "<td class='success' align='center'>0</td>";
        $Withdrawn_Link = "<td class='failures' align='center'><a class='report_link' href='$Report_Path#Withdrawn'>$TestResult{$Library}{'Withdrawn'} absent</a></td>" if($TestResult{$Library}{'Withdrawn'}>0);
        
        my $T_High_Count = $TestResult{$Library}{'T_High'};
        my $T_High_Link = "<td class='success' align='center'>0</td>";
        $T_High_Link = "<td class='failures' align='center'><a class='report_link' href='$Report_Path#Type_Problems_High'>$T_High_Count ".word_tr("problem", $T_High_Count)."</a></td>" if($T_High_Count>0);
        
        my $T_Medium_Count = $TestResult{$Library}{'T_Medium'};
        my $T_Medium_Link = "<td class='success' align='center'>0</td>";
        $T_Medium_Link = "<td class='failures' align='center'><a class='report_link' href='$Report_Path#Type_Problems_Medium'>$T_Medium_Count ".word_tr("problem", $T_Medium_Count)."</a></td>" if($T_Medium_Count>0);
        
        my $T_Low_Count = $TestResult{$Library}{'T_Low'};
        my $T_Low_Link = "<td class='success' align='center'>0</td>";
        $T_Low_Link = "<td class='warnings' align='center'><a class='report_link' href='$Report_Path#Type_Problems_Low'>$T_Low_Count ".word_tr("warning", $T_Low_Count)."</a></td>" if($T_Low_Count>0);
        
        my $I_High_Count = $TestResult{$Library}{'I_High'};
        my $I_High_Link = "<td class='success' align='center'>0</td>";
        $I_High_Link = "<td class='failures' align='center'><a class='report_link' href='$Report_Path#Interface_Problems_High'>$I_High_Count ".word_tr("problem", $I_High_Count)."</a></td>" if($I_High_Count>0);
        
        my $I_Medium_Count = $TestResult{$Library}{'I_Medium'};
        my $I_Medium_Link = "<td class='success' align='center'>0</td>";
        $I_Medium_Link = "<td class='failures' align='center'><a class='report_link' href='$Report_Path#Interface_Problems_Medium'>$I_Medium_Count ".word_tr("problem", $I_Medium_Count)."</a></td>" if($I_Medium_Count>0);
        
        my $I_Low_Count = $TestResult{$Library}{'I_Low'};
        my $I_Low_Link = "<td class='success' align='center'>0</td>";
        $I_Low_Link = "<td class='warnings' align='center'><a class='report_link' href='$Report_Path#Interface_Problems_Low'>$I_Low_Count ".word_tr("warning", $I_Low_Count)."</a></td>" if($I_Low_Count>0);
        
        my $Constants_Count = $TestResult{$Library}{'Constants'};
        my $Constants_Link = "<td class='success' align='center'>0</td>";
        $Constants_Link = "<td class='warnings' align='center'><a class='report_link' href='$Report_Path#Constants'>$Constants_Count ".word_tr("warning", $Constants_Count)."</a></td>" if($Constants_Count>0);
        
        $LIBS_INFO .= "<tr><td class='table_header'><a style='color:Blue;' href=\'$Report_Path\'>$Library</a></td>".$Withdrawn_Link.$T_High_Link.$T_Medium_Link.$T_Low_Link.$I_High_Link.$I_Medium_Link.$I_Low_Link.$Constants_Link."</tr>\n";
    }
    $LIBS_INFO .= "</table>\n";
    
    my $Total_Problems = $Withdrawn_Total+$T_High_Total+$T_Medium_Total+$I_High_Total+$I_Medium_Total;
    my $Total_Warnings = $T_Low_Total+$I_Low_Total+$Constants_Total;
    my $SUMMARY = "<h2 style='margin-bottom:0px;padding-bottom:0px;'>Summary</h2><hr/>\n<table cellpadding='3' border='1' style='border-collapse:collapse;'>\n";
    $SUMMARY .= "<tr><td class='table_header summary_item'>Total libraries checked</td><td align='right' class='summary_item_value'>$Total_Libs</td></tr>";
    $SUMMARY .= "<tr><td class='table_header summary_item'>Total problems found</td><td align='right' class='summary_item_value'>$Total_Problems</td></tr>";
    $SUMMARY .= "<tr><td class='table_header summary_item'>Total warnings (problems with low risk)</td><td align='right' class='summary_item_value'>$Total_Warnings</td></tr>";
    $SUMMARY .= "</table>\n";
    
    my $PROBLEM_SUMMARY = "<h2 style='margin-bottom:0px;padding-bottom:0px;'>Problem Summary</h2><hr/>\n";
    $PROBLEM_SUMMARY .= create_summary($Withdrawn_Total, $T_High_Total, $T_Medium_Total, $T_Low_Total, $I_High_Total, $I_Medium_Total, $I_Low_Total, $Constants_Total);
    
    my $INDEX = "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\">\n<head>\n<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\" />
    <title>ABI compliance report for $Distribution_Name and $TESTSUITE_DISTRIBUTION</title>\n<!--Styles-->\n".$CssStyles."\n<!--Styles_End-->\n</head>\n<body><h1>ABI compliance report for the <span style='color:Blue;white-space:nowrap;'>$Distribution_Name</span> and <span style='color:Blue;white-space:nowrap;'>$TESTSUITE_DISTRIBUTION</span> libraries</h1>\n$SUMMARY\n$PROBLEM_SUMMARY\n$LIBS_INFO\n</body></html>";
    open(INDEX, ">results/abi_compliance_report.html");
    print INDEX $INDEX;
    close(INDEX);
    print "see the ABI compliance report in file 'results/abi_compliance_report.html'\n";
}

sub cmd_cat($$)
{
    my ($Path, $Grep) = @_;
    return "" if(not $Path or not -e $Path);
    my $Cmd = "cat ".esc($Path);
    if($Grep)
    {
        $Cmd .= " | grep \"$Grep\"";
    }
    return `$Cmd`;
}

sub check_result($)
{
    my $Library = $_[0];
    return if(not -e "compat_reports/$Library");
    my $Report_Path = (cmd_find("compat_reports/$Library","f","abi_compat_report.html"))[0];
    return if(not $Report_Path);
    my $ABI_COMPAT_REPORT = cmd_cat($Report_Path, "");
    my ($Withdrawn, $T_High, $T_Medium, $T_Low, $I_High, $I_Medium, $I_Low, $Constants) = (0, 0, 0, 0, 0, 0, 0, 0);
    if($ABI_COMPAT_REPORT =~ /\<input id='Withdrawn_Interfaces_Count' type='hidden' value='([0-9]+)' \/\>/m)
    {
        $Withdrawn = $1;
    }
    if($ABI_COMPAT_REPORT =~ /\<input id='Type_Problems_High_Count' type='hidden' value='([0-9]+)' \/\>/m)
    {
        $T_High = $1;
    }
    if($ABI_COMPAT_REPORT =~ /\<input id='Type_Problems_Medium_Count' type='hidden' value='([0-9]+)' \/\>/m)
    {
        $T_Medium = $1;
    }
    if($ABI_COMPAT_REPORT =~ /\<input id='Type_Problems_Low_Count' type='hidden' value='([0-9]+)' \/\>/m)
    {
        $T_Low = $1;
    }
    if($ABI_COMPAT_REPORT =~ /\<input id='Interface_Problems_High_Count' type='hidden' value='([0-9]+)' \/\>/m)
    {
        $I_High = $1;
    }
    if($ABI_COMPAT_REPORT =~ /\<input id='Interface_Problems_Medium_Count' type='hidden' value='([0-9]+)' \/\>/m)
    {
        $I_Medium = $1;
    }
    if($ABI_COMPAT_REPORT =~ /\<input id='Interface_Problems_Low_Count' type='hidden' value='([0-9]+)' \/\>/m)
    {
        $I_Low = $1;
    }
    if($ABI_COMPAT_REPORT =~ /\<input id='Changed_Constants_Count' type='hidden' value='([0-9]+)' \/\>/m)
    {
        $Constants = $1;
    }
    $TestResult{$Library} = {"Withdrawn"=>$Withdrawn, "T_High"=>$T_High, "T_Medium"=>$T_Medium, "T_Low"=>$T_Low, "I_High"=>$I_High, "I_Medium"=>$I_Medium, "I_Low"=>$I_Low, "Constants"=>$Constants};
}

sub create_summary($$$$$$$$)
{
    my ($Withdrawn, $T_High, $T_Medium, $T_Low, $I_High, $I_Medium, $I_Low, $Constants) = @_;
    my $SUMMARY = "<table cellpadding='3' border='1' style='border-collapse:collapse;'>";
    $SUMMARY .= "<tr><td class='table_header summary_item' colspan='2'>Absent Interfaces<br/>in $TESTSUITE_DISTRIBUTION</td><td align='right' class='summary_item_value'>$Withdrawn</td></tr>";
    
    $SUMMARY .= "<tr><td class='table_header summary_item' colspan='2'>Constants Problems</td><td align='right' class='summary_item_value'>$Constants</td></tr>";
    
    $SUMMARY .= "<tr><td class='table_header summary_item' rowspan='3'>Problems in<br/>Data Types</td><td class='table_header summary_item' style='color:Red;'>High risk</td><td align='right' class='summary_item_value'>$T_High</td></tr>";
    
    $SUMMARY .= "<tr><td class='table_header summary_item'>Medium risk</td><td align='right' class='summary_item_value'>$T_Medium</td></tr>";
    
    $SUMMARY .= "<tr><td class='table_header summary_item'>Warnings</td><td align='right' class='summary_item_value'>$T_Low</td></tr>";
    
    $SUMMARY .= "<tr><td class='table_header summary_item' rowspan='3'>Interface<br/>Problems</td><td class='table_header summary_item' style='color:Red;'>High risk</td><td align='right' class='summary_item_value'>$I_High</td></tr>";
    
    $SUMMARY .= "<tr><td class='table_header summary_item'>Medium risk</td><td align='right' class='summary_item_value'>$I_Medium</td></tr>";
    
    $SUMMARY .= "<tr><td class='table_header summary_item'>Warnings</td><td align='right' class='summary_item_value'>$I_Low</td></tr>";
    $SUMMARY .= "</table>";
    return $SUMMARY;
}

sub readDescriptor($$)
{
    my ($Library, $Descriptor_Path) = @_;
    return if(not $Descriptor_Path or not $Library);
    my $Descriptor_File = cmd_cat($Descriptor_Path, "");
    $Descriptor_File =~ s|/\*[^*]*\*/||igs;
    if(not $Descriptor_File)
    {
        print "WARNING: descriptor $Descriptor_Path is empty\n";
    }
    if($Descriptor_File =~ m/<version>[ \n]*(.*?)[ \n]*<\/version>/ios)
    {
        $Descriptor{$Library}{'Version'} = $1;
    }
    else
    {
        print "WARNING: version is not specified in the $Descriptor_Path\n";
    }
    if($Descriptor_File =~ m/<headers>[ \n]*(.*?)[ \n]*<\/headers>/ios)
    {
        $Descriptor{$Library}{'Headers'} = $1;
    }
    else
    {
        print "WARNING: header files were not specified in the $Descriptor_Path\n";
    }
    if($Descriptor_File =~ m/<libs>[ \n]*(.*?)[ \n]*<\/libs>/ios)
    {
        $Descriptor{$Library}{'Libs'} = $1;
    }
    else
    {
        print "WARNING: shared objects were not specified in the $Descriptor_Path\n";
    }
    if($Descriptor_File =~ m/<include_paths>[ \n]*(.*?)[ \n]*<\/include_paths>/ios)
    {
        $Descriptor{$Library}{'Include_Paths'} = $1;
    }
    if($Descriptor_File =~ m/<gcc_options>[ \n]*(.*?)[ \n]*<\/gcc_options>/ios)
    {
        $Descriptor{$Library}{'Gcc_Options'} = $1;
    }
    if($Descriptor_File =~ m/<opaque_types>[ \n]*(.*?)[ \n]*<\/opaque_types>/ios)
    {
        $Descriptor{$Library}{'Opaque_Types'} = $1;
    }
    if($Descriptor_File =~ m/<skip_interfaces>[ \n]*(.*?)[ \n]*<\/skip_interfaces>/ios)
    {
        $Descriptor{$Library}{'Skip_Interfaces'} = $1;
    }
    if($Descriptor_File =~ m/<include_preamble>[ \n]*(.*?)[ \n]*<\/include_preamble>/ios)
    {
        $Descriptor{$Library}{'Include_Preamble'} = $1;
    }
}

sub create_target_descriptor($$)
{
    `mkdir -p target_descriptors`;
    my ($Library, $Descriptor_Path) = @_;
    readDescriptor($Library, $Descriptor_Path);
    my $TARGET_DESCRIPTOR = "<version>\n    $Distribution_Name\n</version>\n\n";
    my $HEADERS = "";
    foreach my $Header_Path (split("\n", $Descriptor{$Library}{'Headers'}))
    {
        $Header_Path =~ s/\A[ ]+//g;
        $Header_Path =~ s/[ ]+\Z//g;
        my $New_Path = findPath($Header_Path);
        if($New_Path)
        {
            $HEADERS .= "    $New_Path\n";
        }
        else
        {
            print "WARNING: corresponding header file for \'$Header_Path\' was not found\n";
        }
    }
    return 0 if(not $HEADERS);
    $TARGET_DESCRIPTOR .= "<headers>\n$HEADERS</headers>\n\n";
    
    my $LIBS = "";
    foreach my $Lib_Path (split("\n", $Descriptor{$Library}{'Libs'}))
    {
        $Lib_Path =~ s/\A[ ]+//g;
        $Lib_Path =~ s/[ ]+\Z//g;
        my $New_Path = findPath($Lib_Path);
        if($New_Path)
        {
            $LIBS .= "    $New_Path\n";
        }
        else
        {
            print "WARNING: corresponding shared object for \'$Lib_Path\' was not found\n";
        }
    }
    return 0 if(not $LIBS);
    $TARGET_DESCRIPTOR .= "<libs>\n$LIBS</libs>\n\n";
    
    my $INCLUDE_PATHS = "";
    foreach my $Include_Path (split("\n", $Descriptor{$Library}{'Include_Paths'}))
    {
        $Include_Path =~ s/\A[ ]+//g;
        $Include_Path =~ s/[ ]+\Z//g;
        my $New_Path = findPath($Include_Path);
        if($New_Path)
        {
            $INCLUDE_PATHS .= "    $New_Path\n";
        }
        else
        {
            print "WARNING: corresponding include path for \'$Include_Path\' was not found\n";
        }
    }
    return 0 if(not $INCLUDE_PATHS and $Descriptor{$Library}{'Include_Paths'});
    $TARGET_DESCRIPTOR .= $INCLUDE_PATHS;
    $TARGET_DESCRIPTOR .= "<include_paths>\n$INCLUDE_PATHS</include_paths>\n\n" if($INCLUDE_PATHS);
    
    $TARGET_DESCRIPTOR .= "<gcc_options>\n    $Descriptor{$Library}{'Gcc_Options'}\n</gcc_options>\n\n" if($Descriptor{$Library}{'Gcc_Options'});
    $TARGET_DESCRIPTOR .= "<opaque_types>\n    $Descriptor{$Library}{'Opaque_Types'}\n</opaque_types>\n\n" if($Descriptor{$Library}{'Opaque_Types'});
    $TARGET_DESCRIPTOR .= "<skip_interfaces>\n    $Descriptor{$Library}{'Skip_Interfaces'}\n</skip_interfaces>\n\n" if($Descriptor{$Library}{'Skip_Interfaces'});
    $TARGET_DESCRIPTOR .= "<include_preamble>\n    $Descriptor{$Library}{'Include_Preamble'}\n</include_preamble>\n\n" if($Descriptor{$Library}{'Include_Preamble'});
    
    open(TARGET_DESCRIPTOR, ">target_descriptors/$Library\.desc");
    print TARGET_DESCRIPTOR $TARGET_DESCRIPTOR;
    close(TARGET_DESCRIPTOR);
    return 1;
}

sub findPath($)
{
    my $Path = $_[0];
    if($Path !~ /\// or -e $Path)
    {
        return $Path;
    }
    else
    {
        findDevelFiles_1stStage();
        my $NewPath = searchPath($Path);
        if($NewPath)
        {
            return $NewPath;
        }
    }
    return "";
}



sub findDevelFiles_1stStage()
{
    return if(keys(%DevelFiles));
    foreach my $Dest (sort {lc($a) cmp lc($b)} split("\n", `find /usr/include /usr/lib /lib`))
    {
        chomp($Dest);
        $DevelFiles{$Dest} = 1;
    }
}

sub searchPath($)
{
    my $Path = $_[0];
    $Path =~ s!\.so\.[0-9.]*\Z!.so!g;
    while($Path=~s!\A(.+?)\/!!o)
    {
        foreach my $DevelPath (keys(%DevelFiles))
        {
            if($DevelPath=~/\/$Path[0-9.]*\Z/)
            {
                return $DevelPath;
            }
        }
    }
}

sub detect_OS()
{
    my $Distribution = `lsb_release -s -d`;
    chomp($Distribution);
    $Distribution =~ s/^\s*\"(.*)\"\s*$/$1/;
    return $Distribution;
}

sub scenario()
{
    HELP_MESSAGE() if (defined $Help);
    if(not $Distribution_Name)
    {
        print "WARNING: distribution name was not selected\n";
        print "selecting ... ";
        $Distribution_Name = detect_OS();
        if($Distribution_Name)
        {
            print $Distribution_Name."\n";
        }
        else
        {
            print "can't select\n";
            exit(1);
        }
    }
    if(defined $Create_TestSuite)
    {
        if(not $Source)
        {
            print "select test suite source path (-src <path> option)\n";
            exit(1);
        }
        if(not -e $FindBin::Bin."/abi-compliance-checker.pl")
        {
            print "'abi-compliance-checker.pl' was not found\n";
            exit(1);
        }
        if(not -e $Source."/int_list/")
        {
            print "ERROR: interface lists were not found in the directory '$Source/int_list/'\n";
            exit(1);
        }
        if(not -e $Source."/descriptors/")
        {
            print "ERROR: library descriptors were not found in the directory '$Source/descriptors/'\n";
            exit(1);
        }
        if($TargetLibrary)
        {
            recreate_test_case($TargetLibrary);
        }
        else
        {
            create_test_suite();
        }
    }
    if(defined $Run_Tests)
    {
        if(not $Source)
        {
            print "select test suite source path (-src <path> option)\n";
            exit(1);
        }
        if(not -e $Source."/distribution.name")
        {
            print "'$Source/distribution.name' is not exists\n";
            exit(1);
        }
        if(not -e $FindBin::Bin."/abi-compliance-checker.pl")
        {
            print "'abi-compliance-checker.pl' was not found\n";
            exit(1);
        }
        if(not -e $Source."/int_list/")
        {
            print "ERROR: interfaces lists were not found in the directory '$Source/int_list/'\n";
            exit(1);
        }
        if(not -e $Source."/descriptors/")
        {
            print "ERROR: library descriptors were not found in the directory '$Source/descriptors/'\n";
            exit(1);
        }
        if(not -e $Source."/test_data/")
        {
            print "ERROR: test data was not found in the directory '$Source/test_data/'\n";
            exit(1);
        }
        $TESTSUITE_DISTRIBUTION = cmd_cat($Source."/distribution.name", "");
        if(not $TESTSUITE_DISTRIBUTION)
        {
            print "distribution name is not set in the file '$Source/distribution.name'\n";
            exit(1);
        }
        `rm -fr results/ && mkdir -p results/`;
        run_tests();
        create_index();
        `mv compat_reports results/compat_reports`;
    }
}

#SCENARIO
scenario();
